
/* include the library headers */
#include <lcrzo.h>
#include "lcrzoex.h"
#include "shared/shared_misc.h"

/*-------------------------------------------------------------*/
/* set information about this example */
lcrzoex_infos lcrzoex_000244=
{ /*reference*/     "lcrzoex", 244,
  /*version*/       1, 0,
  /*title*/         "udp traceroute (spoofed)",
  /*french title*/  "traceroute udp (spoofe)",
  /*description*/   "",
  /*french desc.*/  "",
  /*usage*/         "device srceth routereth srcip dstip dstport maxttl",
  /*french usage*/  "device ethsrc ethrouteur ipsrc ipdst portdst ttlmax",
  /*usage example*/ "eth0 aa:bb:cc:dd:ee:ff 11:22:33:44:55:66 1.2.3.4 5.6.7.78 53 30",
  /*fr. usage ex.*/ "eth0 aa:bb:cc:dd:ee:ff 11:22:33:44:55:66 1.2.3.4 5.6.7.78 53 30",
  /*author*/        "Laurent"
};

/*-------------------------------------------------------------*/
/* function declarations */
int lcrzoex_000244_main(int argc, char *argv[]);
void lcrzoex_000244_sig_received1(int i);
void lcrzoex_000244_sig_received2(int i);
int lcrzoex_000244_traceroute(lcrzo_device device,
			      lcrzo_etha clientetha,
			      lcrzo_etha routeretha,
			      lcrzo_ipl clientipl, 
			      lcrzo_ipl serveripl, 
			      lcrzo_uint16 serverport,
			      lcrzo_uint8 maxttl);
int lcrzoex_000244_sendudp(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_etha routeretha,
			   lcrzo_ipl clientipl, 
			   lcrzo_ipl serveripl, 
			   lcrzo_uint16 serverport,
			   lcrzo_uint16 startingclientport,
			   lcrzo_uint8 maxttl);
int lcrzoex_000244_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_etha routeretha,
			 lcrzo_ipl clientipl, 
			 lcrzo_ipl serveripl, 
			 lcrzo_uint16 serverport,
			 lcrzo_uint16 startingclientport,
			 lcrzo_uint8 maxttl);
int lcrzoex_000244_printroutei(lcrzo_etha clientetha,
			       lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 serverport,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint8 maxttl,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize);

/*-------------------------------------------------------------*/
/* global variables */
int lcrzoex_000244_pid;
lcrzo_spoof lcrzoex_000244_spoof;
lcrzo_sniff lcrzoex_000244_sniff;

/*-------------------------------------------------------------*/
/* main function : you can rename it to "main", and then this file
   can be compiled alone. */
int lcrzoex_000244_main(int argc, char *argv[])
{ lcrzo_device device;
  lcrzo_etha clientetha, routeretha;
  lcrzo_ipl clientipl, serveripl;
  lcrzo_uint16 serverport;
  lcrzo_uint8 maxttl;

  if ( argc<8 )
  { /* print usage */
    lcrzo_epr(lcrzoex_infos_print(lcrzoex_000244, argv[0]));
    return(LCRZO_ERR_BULCRZOEXBADUSAGE);
  }
  lcrzo_epr(lcrzo_device_init(argv[1], device));
  lcrzo_epr(lcrzo_etha_init_eths(argv[2], clientetha));
  lcrzo_epr(lcrzo_etha_init_eths(argv[3], routeretha));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[4], &clientipl));
  lcrzo_epr(lcrzo_ipl_init_hs(argv[5], &serveripl));
  serverport=(lcrzo_uint16)atoi(argv[6]);
  maxttl=(lcrzo_uint8)atoi(argv[7]); if (maxttl==0) maxttl=1;

  /* call the main function */
  lcrzo_epr(lcrzoex_000244_traceroute(device, clientetha, routeretha,
				      clientipl, serveripl, serverport,
				      maxttl));

  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000244_traceroute(lcrzo_device device,
			      lcrzo_etha clientetha,
			      lcrzo_etha routeretha,
			      lcrzo_ipl clientipl, 
			      lcrzo_ipl serveripl, 
			      lcrzo_uint16 serverport,
			      lcrzo_uint8 maxttl)
{ lcrzo_uint16 startingclientport;

  /* determine starting values */
  lcrzo_epr(lcrzo_uint16_rand(5000, 30000, &startingclientport));

  lcrzoex_000244_pid=fork();
  if (lcrzoex_000244_pid<0) return(LCRZO_ERR_FUFORK);
  if (lcrzoex_000244_pid==0)
  { /* this process sends UDP packets */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000244_sig_received1);
    lcrzo_epr(lcrzoex_000244_sendudp(device, clientetha, routeretha,
				     clientipl, serveripl, serverport,
				     startingclientport, maxttl));
  }
  else
  { /* this process sniff and prints IP addresses of routers */
    /* in case we receive a signal */
    signal(SIGINT, lcrzoex_000244_sig_received2);
    lcrzo_epr(lcrzoex_000244_print(device, clientetha, routeretha,
				   clientipl, serveripl, serverport,
				   startingclientport, maxttl));
  }
  return (LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
void lcrzoex_000244_sig_received1(int i)
{ int language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_FRLANG )
    printf("Fermeture du spoof (recu : SIG%d)\n", i);
  else
    printf("Closing spoof (received : SIG%d)\n", i);
  lcrzo_spoof_close(&lcrzoex_000244_spoof);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
void lcrzoex_000244_sig_received2(int i)
{ int status, language;

  lcrzo_global_get_language(&language);
  if ( language==LCRZO_GLOBAL_FRLANG )
    printf("Fermeture du sniff (recu : SIG%d)\n", i);
  else
    printf("Closing sniff (received : SIG%d)\n", i);
  kill(lcrzoex_000244_pid, SIGINT);
  waitpid(lcrzoex_000244_pid, &status, 0);
  lcrzo_sniff_close(&lcrzoex_000244_sniff);
  raise(SIGTERM);
}

/*-------------------------------------------------------------*/
int lcrzoex_000244_sendudp(lcrzo_device device,
			   lcrzo_etha clientetha,
			   lcrzo_etha routeretha,
			   lcrzo_ipl clientipl, 
			   lcrzo_ipl serveripl, 
			   lcrzo_uint16 serverport,
			   lcrzo_uint16 startingclientport,
			   lcrzo_uint8 maxttl)
{ lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip;
  lcrzo_hdrludp hdrludp;
  lcrzo_uint8 i;

  /* initialize Ethernet header */
  lcrzo_epr(lcrzo_hdrleth_initdefault(&hdrleth));
  memcpy(hdrleth.src, clientetha, LCRZO_ETHA_MAXBYTES);
  memcpy(hdrleth.dst, routeretha, LCRZO_ETHA_MAXBYTES);

  /* initialize IP header */
  lcrzo_epr(lcrzo_hdrlip_initdefault(&hdrlip));
  hdrlip.saddr=clientipl;
  hdrlip.daddr=serveripl;
  hdrlip.ttl=1;
  hdrlip.id=startingclientport;

  /* initialize UDP header */
  lcrzo_epr(lcrzo_hdrludp_initdefault(&hdrludp));
  hdrludp.sport=startingclientport;
  hdrludp.dport=serverport;

  /* spoof */
  lcrzo_epr(lcrzo_spoof_init(&lcrzoex_000244_spoof));
  /*lcrzo_epr(lcrzo_spoof_set_printbeforesending(&lcrzoex_000244_spoof,
                                               LCRZO_TRUE));*/
  for ( i=1 ; i<=maxttl ; i++ )
  { lcrzo_epr(lcrzo_spoof_ethipoptudpdata(&lcrzoex_000244_spoof,
					     device, hdrleth,
					     hdrlip, NULL, 0,
					     hdrludp, NULL, 0));
    hdrlip.id++;
    hdrlip.ttl++;
    hdrludp.sport++;
    lcrzo_epr(lcrzo_time_sleep(1000000));
  }
  lcrzo_epr(lcrzo_spoof_close(&lcrzoex_000244_spoof));
  
  return(LCRZO_ERR_OK);
}

/*-------------------------------------------------------------*/
int lcrzoex_000244_print(lcrzo_device device,
			 lcrzo_etha clientetha,
			 lcrzo_etha routeretha,
			 lcrzo_ipl clientipl, 
			 lcrzo_ipl serveripl, 
			 lcrzo_uint16 serverport,
			 lcrzo_uint16 startingclientport,
			 lcrzo_uint8 maxttl)
{ lcrzo_uint32 mtu;
  lcrzo_data sniffdata;
  lcrzo_int32 sniffdatasize;
  int ret;

  /* init sniff */
  lcrzo_epr(lcrzo_mtu_init_device(device, &mtu));
  lcrzo_epr(lcrzo_sniff_init(device, mtu, "", &lcrzoex_000244_sniff));

  /* main loop */
  ret=lcrzo_sniff_nextm(&lcrzoex_000244_sniff, LCRZO_TRUE, 
			LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  while (ret==LCRZO_ERR_OK)
  { /* answer to requests */
    lcrzo_epr(lcrzo_virtual_answer_alive(sniffdata, sniffdatasize, 
                                         device, clientetha, clientipl));
    /* decode an icmp error, and print it */
    lcrzo_epr(lcrzoex_000244_printroutei(clientetha, routeretha,
					 clientipl, serveripl, serverport,
					 startingclientport, maxttl,
					 sniffdata, sniffdatasize));
    lcrzo_data_free(sniffdata);
    /* sniff next packet */
    ret=lcrzo_sniff_nextm(&lcrzoex_000244_sniff, LCRZO_TRUE,
			  LCRZO_SNIFF_TYPE_NORMAL, &sniffdata, &sniffdatasize);
  }

  /* close everything */
  raise(SIGINT);
  /*never reatched*/
  return(LCRZO_ERR_IEINTERNALERROR);
}

/*-------------------------------------------------------------*/
int lcrzoex_000244_printroutei(lcrzo_etha clientetha,
			       lcrzo_etha routeretha,
			       lcrzo_ipl clientipl, 
			       lcrzo_ipl serveripl, 
			       lcrzo_uint16 serverport,
			       lcrzo_uint16 startingclientport,
			       lcrzo_uint8 maxttl,
			       lcrzo_data sniffdata,
			       lcrzo_int32 sniffdatasize)
{ static int ttlreatched=0xFFFF; 
  lcrzo_hdrleth hdrleth;
  lcrzo_hdrlip hdrlip, badhdrip;
  lcrzo_data baddata;
  lcrzo_string string;
  lcrzo_uint16 baddatasize, badudpsport, badudpdport;
  lcrzo_uint8 type, code;
  int ret, indexidok, indexidko, indexsport, language;

  /*lcrzo_epr(lcrzo_packet_print_eth(sniffdata, sniffdatasize,
                                   LCRZO_PRINTPROFILE_ARRAYDUMP));*/

  /* try to decode icmp */
  ret=lcrzo_packet_decodem_ethipopticmperror(sniffdata, sniffdatasize,
					     &hdrleth, &hdrlip, NULL,NULL,
					     &type, &code, NULL, NULL,
					     &baddata, &baddatasize);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check if the addresses are the one we are looking for */
  if ( ! lcrzo_etha_equal(hdrleth.src, routeretha) ||
       ! lcrzo_etha_equal(hdrleth.dst, clientetha) ||
       ! lcrzo_ipl_equal(hdrlip.daddr, clientipl) )
  { lcrzo_data_free(baddata);
    return(LCRZO_ERR_OK);
  }

  /* icmp should contain udp */
  ret=lcrzo_ip64bits_decodem_ipudp(baddata, baddatasize,
				   &badhdrip, &badudpsport, &badudpdport,
				   NULL, NULL);
  lcrzo_data_free(baddata);
  if ( ret!=LCRZO_ERR_OK )
    return(LCRZO_ERR_OK);

  /* check various values in baddata */
  if ( badudpdport!=serverport ||
       badhdrip.saddr!=clientipl ||
       badhdrip.daddr!=serveripl )
    return(LCRZO_ERR_OK);

  /* compute indexid from the id of the received packet */
  indexidok=badhdrip.id-startingclientport+1;
  /* this one is because there is a bug in the IP stack of BSD :
     id is reversed :( */
  indexidko= ( (badhdrip.id&0xFF)<<8 ) | ( (badhdrip.id&0xFF00)>>8 );
  indexidko=indexidko-startingclientport+1;
  /* compute indexsport from the sport of the received packet */
  indexsport=badudpsport-startingclientport+1;

  /* verify that indexes are equal */
  if ( indexidok!=indexsport && indexidko!=indexsport )
    return(LCRZO_ERR_OK);

  /* print the result */
  if ( indexsport<ttlreatched )
  { printf("% 3d : ", indexsport);
    lcrzo_epr(lcrzo_ipl_print("", hdrlip.saddr, ""));
  
    /* print the type of error */
    if ( type==LCRZO_ICMPTYPE_TIMEEXCEED &&
	 code == LCRZO_ICMPCODE_TIMEEXCEED_TTL )
    { /* do not display the error type and code */
      printf("\n"); 
    }
    else if ( type==LCRZO_ICMPTYPE_DESTUNREACH &&
	      code == LCRZO_ICMPCODE_DESTUNREACH_PORT )
    { /* host reatched */
      printf("\n"); 
      ttlreatched=indexsport;
      lcrzo_epr(lcrzo_global_get_language(&language));
      if ( language==LCRZO_GLOBAL_FRLANG )
      { printf("Destination atteinte.\n");
        printf("Note : le port UDP %d est ferme (on a recu un ICMP).\n", 
	       serverport);
        printf("On se donne 3 secondes pour attendre les eventuelles ");
        printf("reponses retardataires.\n");
      }
      else
      { printf("Destination reatched.\n");
        printf("Note : the UDP port %d is closed (an ICMP was received).\n", 
	       serverport);
        printf("We continue during 3 seconds to wait for late answers.\n");
      }
      signal(SIGALRM, lcrzoex_000244_sig_received2);
      alarm(3);
    }
    else 
    { /* display the error type and code */
      lcrzo_epr(lcrzo_string_initm_icmptypecode(type, code, &string));
      printf(" : %s\n", string);
      lcrzo_string_free(string);
    }

    /* if maxttl was too short */
    if ( indexsport == maxttl )
    { lcrzo_epr(lcrzo_global_get_language(&language));
      if ( language==LCRZO_GLOBAL_FRLANG )
	printf("maxttl(%d) est trop petit pour atteindre la destination\n", 
	       maxttl);
      else
	printf("maxttl(%d) was too short to reach the destination\n", 
	       maxttl);
      raise(SIGINT);
    }
  }

  return(LCRZO_ERR_OK);
}








